"""
Basket class for calculating the returns and carries of baskets of financial contracts
using various weighting methods.
"""

import numpy as np
import pandas as pd
import seaborn as sns
import warnings

import random
from typing import List, Union, Tuple, Dict
from macrosynergy.panel.historic_vol import expo_weights, expo_std, flat_std
from macrosynergy.management.utils import reduce_df_by_ticker
from macrosynergy.panel.converge_row import ConvergeRow
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.management.simulate import make_qdf
import matplotlib.pyplot as plt
import matplotlib.dates as mdates


class Basket(object):
    """
    Calculates the returns and carries of baskets of financial contracts using various
    weighting methods.

    Parameters
    ----------
    df : ~pandas.Dataframe
        standardized DataFrame containing the columns: 'cid', 'xcat', 'real_date' and
        'value'.
    contracts : List[str]
        base tickers (combinations of cross-sections and base categories) that define
        the contracts that go into the basket.
    ret : str
        return category postfix to be appended to the contract base; default is
        "XR_NSA".
    cry : List[str] or str
        carry category postfix; default is None. The field can either be a single carry
        or multiple carries defined in a List.
    start : str
        earliest date in ISO 8601 format. Default is None.
    end : str
        latest date in ISO 8601 format. Default is None.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the DataFrame. If
        one cross-section has several blacklist periods append numbers to the cross-section
        code.
    ewgts : List[str]
        one or more postfixes that may identify exogenous weight categories. Similar to
        return postfixes they are appended to base tickers.


    .. note::
        Each instance of the class will update associated standardised DataFrames,
        containing return and carry categories, and external weights.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        contracts: List[str],
        ret: str = "XR_NSA",
        cry: Union[str, List[str]] = None,
        start: str = None,
        end: str = None,
        blacklist: dict = None,
        ewgts: List[str] = None,
    ):
        df = QuantamentalDataFrame(df[["cid", "xcat", "real_date", "value"]])
        self._as_categorical = df.InitializedAsCategorical
        assert isinstance(contracts, list)
        c_error = "Contracts must be a list of strings."
        assert all(isinstance(c, str) for c in contracts), c_error
        assert isinstance(ret, str), "`ret` must be a string"

        df = reduce_df_by_ticker(
            df, start=start, end=end, ticks=None, blacklist=blacklist
        )
        self.contracts = contracts
        self.ret = ret
        self.ticks_ret = [con + ret for con in contracts]
        dfw_ret = self.pivot_dataframe(df, self.ticks_ret)
        self.dfw_ret = dfw_ret.dropna(axis=0, how="all")

        self.store_attributes(df, cry, "cry")
        self.store_attributes(df, ewgts, "wgt")

        self.tickers = self.ticks_ret + self.ticks_cry + self.ticks_wgt
        self.start = self.date_check(start)
        self.end = self.date_check(end)
        self.dfx = reduce_df_by_ticker(df, ticks=self.tickers)
        self.dict_retcry = {}  # dictionary for collecting basket return/carry dfs.
        self.dict_wgs = {}  # dictionary for collecting basket return/carry dfs.

    def store_attributes(self, df: pd.DataFrame, pfx: List[str], pf_name: str):
        """
        Adds multiple attributes to class based on postfixes that denote carry or
        external weight types.

        Parameters
        ----------
        df : ~pandas.DataFrame
            original, standardised DataFrame.
        pfx : List[str]
            category postfixes involved in the basket calculation.
        pf_name : str
            associated name of the postfix "cry" or "wgt".  Note: These are [1] flags of
            existence of carry and weight strings in class, [2] lists of tickers related to
            all postfixes, [3] a dictionary of wide time series panel dataframes for all
            postfixes.
        """

        pfx_flag = pfx is not None
        self.__dict__[pf_name + "_flag"] = pfx_flag
        self.__dict__["ticks_" + pf_name] = []
        if pfx_flag:
            error = f"'{pf_name}' must be a <str> or a <List[str]>."
            assert isinstance(pfx, (list, str)), error
            pfx = [pfx] if isinstance(pfx, str) else pfx
            self.__dict__[pf_name] = pfx

            dfws_pfx = {}
            for cat in pfx:
                ticks = [con + cat for con in self.contracts]
                self.__dict__["ticks_" + pf_name] += ticks
                dfws_pfx[cat] = self.pivot_dataframe(df, ticks)
        else:
            dfws_pfx = None

        self.__dict__["dfws_" + pf_name] = dfws_pfx

    @staticmethod
    def pivot_dataframe(df: QuantamentalDataFrame, tick_list: List[str]):
        """
        Reduces the standardised DataFrame to include a subset of the possible tickers
        and, subsequently returns a wide dataframe: each column corresponds to a ticker.

        Parameters
        ----------
        tick_list : List[str]
            list of the respective tickers.
        df : ~pandas.DataFrame
            standardised dataframe.

        Returns
        -------
        ~pandas.DataFrame
            wide dataframe.
        """

        return (
            QuantamentalDataFrame(df=df)
            .reduce_df_by_ticker(tickers=tick_list)
            .to_wide()
        )

    @staticmethod
    def date_check(date_string):
        """
        Validates that the dates passed are valid timestamp expressions and will convert
        to the required form '%Y-%m-%d'. Will raise an assertion if not in the expected
        form.

        Parameters
        ----------
        date_string : str
            valid date expression. For instance, "1st January, 2000."
        """

        date_error = "Expected form of string: '%Y-%m-%d'."
        if date_string is not None:
            try:
                pd.Timestamp(date_string).strftime("%Y-%m-%d")
            except ValueError:
                raise AssertionError(date_error)
            else:
                return pd.Timestamp(date_string).strftime("%Y-%m-%d")

    @staticmethod
    def check_weights(weight: pd.DataFrame):
        """
        Checks if all rows in dataframe add up to roughly 1.

        Parameters
        ----------
        weight : ~pandas.DataFrame
            weight dataframe.
        """

        check = weight.sum(axis=1)
        c = ~((abs(check - 1) < 1e-6) | (abs(check) < 1e-6))
        assert not any(c), f"weights must sum to one (or zero), not: {check[c]}"

    @staticmethod
    def max_weight_func(weights: pd.DataFrame, max_weight: float):
        """
        Enforces maximum weight caps or - if impossible applies equal weight.

        Parameters
        ----------
        weights : ~pandas.DataFrame
            Corresponding weight matrix. Multidimensional.
        max_weight : float
            Upper-bound on the weight allowed for each cross-section.

        Returns
        -------
        ~pandas.DataFrame
            Will return the modified weight DataFrame.


        .. note::
            If the maximum weight is
            less than the equal weight weight, this replaces the computed weight with the
            equal weight. For instance, [np.nan, 0.63, np.nan, np.nan, 0.27] becomes
            [np.nan, 0.5, np.nan, np.nan, 0.5]. Otherwise, the function calls the
            ConvergeRow Class to ensure all weights "converge" to a value within the upper-
            bound. Allow for a margin of error set to 0.001.
        """

        dfw_wgs = weights.to_numpy()

        for i, row in enumerate(dfw_wgs):
            row = ConvergeRow.application(row, max_weight)
            weights.iloc[i, :] = row

        return weights

    def equal_weight(self, df_ret: pd.DataFrame) -> pd.DataFrame:
        """
        Calculates dataframe of equal weights based on available return data.

        Parameters
        ----------
        df_ret : ~pandas.DataFrame
            wide time-indexed data frame of returns.

        Returns
        -------
        ~pandas.DataFrame
            dataframe of weights.


        Notes
        -----
        The method determines the number of non-NA cross-sections per timestamp, and
        subsequently distributes the weights evenly across non-NA cross-sections.
        """

        act_cross = ~df_ret.isnull()
        uniform = (1 / act_cross.sum(axis=1)).values
        uniform = uniform[:, np.newaxis]

        broadcast = np.repeat(uniform, df_ret.shape[1], axis=1)

        weight = act_cross.multiply(broadcast)
        cols = weight.columns
        # Zeroes treated as NaNs.
        weight[cols] = weight[cols].replace({0: np.nan})

        self.check_weights(weight=weight)

        return weight

    def fixed_weight(self, df_ret: pd.DataFrame, weights: List[float]):
        """
        Calculates fixed weights based on a single list of values and a corresponding
        return panel dataframe.

        Parameters
        ----------
        df_ret : ~pandas.DataFrame
            Return series matrix. Multidimensional.
        weights : List[float]
            List of floats determining weight allocation.

        Returns
        -------
        ~pandas.DataFrame
            panel of weights
        """

        act_cross = ~df_ret.isnull()

        weights = np.array(weights, dtype=np.float32)
        rows = act_cross.shape[0]
        broadcast = np.tile(weights, (rows, 1))  # Constructs Array by row repetition.

        # Replaces weight factors with zeroes if concurrent return unavailable.
        weight = act_cross.multiply(broadcast)
        cols = weight.columns
        # Zeroes treated as NaNs.
        weight[cols] = weight[cols].replace({0: np.nan})

        weight_arr = weight.to_numpy()  # convert df to np array.
        weight[weight.columns] = weight_arr / np.sum(weight_arr, axis=1)[:, np.newaxis]
        self.check_weights(weight)

        return weight

    def inverse_weight(
        self,
        dfw_ret: pd.DataFrame,
        lback_meth: str = "xma",
        lback_periods: int = 21,
        remove_zeros: bool = True,
    ):
        """
        Calculates weights inversely proportionate to recent return standard deviations.

        Parameters
        ----------
        dfw_ret : ~pandas.DataFrame
            panel dataframe of returns.
        lback_meth : str
            Lookback method for "invsd" weighting method. Default is "xma".
        lback_periods : int
            Lookback periods. Default is 21.  Half-time for "xma" and full lookback
            period for "ma".
        remove_zeros : Bool
            Any returns that are exact zeros will not be included in the lookback window
            and prior non-zero values are added to the window instead.

        Returns
        -------
        ~pandas.DataFrame
            Dataframe of weights.


        .. note::
            The rolling standard deviation will be calculated either using the standard
            moving average (ma) or the exponential moving average (xma). Both will require
            returns before a first weight can be computed.
        """

        if lback_meth == "ma":
            dfwa = dfw_ret.rolling(window=lback_periods).agg(flat_std, remove_zeros)
            dfwa *= np.sqrt(252)

        else:
            half_life = lback_periods
            weights = expo_weights(lback_periods * 2, half_life)
            dfwa = dfw_ret.rolling(window=lback_periods * 2).agg(
                expo_std, w=weights, remove_zeros=remove_zeros
            )

        cols = dfwa.columns
        # Zeroes treated as NaNs.
        dfwa[cols] = dfwa[cols].replace({0: np.nan})

        df_isd = 1 / dfwa
        df_wgts = df_isd / df_isd.sum(axis=1).values[:, np.newaxis]
        self.check_weights(df_wgts)

        return df_wgts

    def values_weight(
        self, dfw_ret: pd.DataFrame, dfw_wgt: pd.DataFrame, weight_meth: str
    ):
        """
        Returns weights based on an external weighting category.

        Parameters
        ----------
        dfw_ret : ~pandas.DataFrame
            Standard wide dataframe of returns across time and contracts.
        dfw_wgt : ~pandas.DataFrame
            Standard wide dataframe of weight category values across time and contracts.
        weight_meth : str


        Returns
        -------
        ~pandas.DataFrame
            Dataframe of weights.
        """

        negative_condition = np.any((dfw_wgt < 0).to_numpy())
        if negative_condition:
            dfw_wgt[dfw_wgt < 0] = 0.0
            warnings.warn("Negative values in the weight matrix set to zero.")

        exo_array = dfw_wgt.to_numpy()
        df_bool = ~dfw_ret.isnull()

        weights_df = df_bool.multiply(exo_array)
        cols = weights_df.columns

        # Zeroes treated as NaNs.
        weights_df[cols] = weights_df[cols].replace({0: np.nan})

        if weight_meth != "values":
            weights_df = 1 / weights_df

        weights = weights_df.divide(weights_df.sum(axis=1), axis=0)
        self.check_weights(weights)

        return weights

    def make_weights(
        self,
        weight_meth: str = "equal",
        weights: List[float] = None,
        lback_meth: str = "xma",
        lback_periods: int = 21,
        ewgt: str = None,
        max_weight: float = 1.0,
        remove_zeros: bool = True,
    ):
        """
        Returns wide dataframe of weights to be used for basket series.

        Parameters
        ----------
        weight_meth : str
            method used for weighting constituent returns and carry. The parameter can
            receive either a single weight method or multiple weighting methods. See
            `make_basket` docstring.
        weights : List[float]
            single list of weights corresponding to the base tickers in `contracts`
            argument. This is only relevant for the fixed weight method.
        lback_meth : str
            look-back method for "invsd" weighting method. Default is Exponential MA,
            "ema". The alternative is simple moving average, "ma".
        lback_periods : int
            look-back periods for "invsd" weighting method. Default is 21.  Half-time
            for "xma" and full lookback period for "ma".
        ewgt : str
            Exogenous weight postfix that defines the weight value panel. Only needed
            for the 'values' or 'inv_values' method.
        max_weight : float
            maximum weight of a single contract. Default is 1, i.e zero restrictions.
            The purpose of the restriction is to limit concentration within the basket.
        remove_zeros : bool
            removes the zeros. Default is set to True.

        Returns
        -------
        pd.DataFrame
            wide dataframe of contract weights across time.
        """

        assert 0.0 < max_weight <= 1.0
        assert weight_meth in ["equal", "fixed", "values", "inv_values", "invsd"]

        # Apply weight method.

        if weight_meth == "equal":
            dfw_wgs = self.equal_weight(df_ret=self.dfw_ret)

        elif weight_meth == "fixed":
            message = "Expects a list of weights."
            message_2 = "List of weights must be equal to the number of contracts."
            assert isinstance(weights, list), message
            assert self.dfw_ret.shape[1] == len(weights), message_2
            message_3 = "Expects a list of floating point values."
            assert all(isinstance(w, (int, float)) for w in weights), message_3

            dfw_wgs = self.fixed_weight(df_ret=self.dfw_ret, weights=weights)

        elif weight_meth == "invsd":
            error_message = "Lookback method method must be 'ma' or 'xma'."
            assert lback_meth in ["xma", "ma"], error_message
            assert isinstance(lback_periods, int), "Expects <int>."
            dfw_wgs = self.inverse_weight(
                dfw_ret=self.dfw_ret,
                lback_meth=lback_meth,
                lback_periods=lback_periods,
                remove_zeros=remove_zeros,
            )

        elif weight_meth in ["values", "inv_values"]:
            assert ewgt in self.wgt, f"{ewgt} is not defined on the instance."
            # Lag by one day to be used as weights.
            try:
                dfw_wgt = self.dfws_wgt[ewgt].shift(1)
            except KeyError as e:
                print(f"Basket not found: {e}.")
            else:
                cols = sorted(dfw_wgt.columns)
                dfw_ret = dfw_wgt.reindex(cols, axis=1)
                dfw_wgt = dfw_wgt.reindex(cols, axis=1)
                dfw_wgs = self.values_weight(dfw_ret, dfw_wgt, weight_meth)

        else:
            raise NotImplementedError(f"Weight method unknown {weight_meth}")

        # Remove leading NA rows.

        fvi = max(dfw_wgs.first_valid_index(), self.dfw_ret.first_valid_index())
        dfw_wgs = dfw_wgs[fvi:]

        # Impose cap on cross-section weights.

        if max_weight < 1.0:
            dfw_wgs = self.max_weight_func(weights=dfw_wgs, max_weight=max_weight)

        return dfw_wgs

    @staticmethod
    def column_manager(df_cat: pd.DataFrame, dfw_wgs: pd.DataFrame):
        """
        Will match the column names of the two dataframes involved in the computation:
        either the return & weight dataframes or the carry & weight dataframes. The
        pandas multiply operation requires the column names, of both dataframes involved
        in the binary operation, to be identical.

        Parameters
        ----------
        df_cat : ~pandas.DataFrame
            return or carry dataframe.
        dfw_wgs : ~pandas.DataFrame
            weight dataframe.

        Returns
        -------
        ~pandas.DataFrame
            modified weight dataframe (column names will map to the other dataframe
            received).
        """

        df_cat = df_cat.reindex(sorted(df_cat.columns), axis=1)
        dfw_wgs = dfw_wgs.reindex(sorted(dfw_wgs.columns), axis=1)

        ret_cols = df_cat.columns
        weight_cols = dfw_wgs.columns

        if all(ret_cols != weight_cols):
            dfw_wgs.columns = ret_cols

        return dfw_wgs

    def column_weights(self, dfw_wgs: pd.DataFrame):
        """
        The weight dataframe is used to compute the basket performance for returns,
        carries etc. Therefore, with their broad application, the column names of the
        dataframe should correspond to the ticker postfix of each contract.

        Parameters
        ----------
        dfw_wgs : ~pandas.DataFrame
            weight dataframe.

        Returns
        -------
        ~pandas.DataFrame
            weight dataframe with updated columns names.
        """

        dfw_weight_names = lambda w_name: w_name[: w_name.find(self.w_field)]
        if self.wgt_flag and self.exo_w_postfix is not None:
            self.w_field = self.exo_w_postfix
        else:
            self.w_field = self.ret

        cols = list(map(dfw_weight_names, dfw_wgs.columns))
        dfw_wgs.columns = cols
        dfw_wgs.columns.name = "ticker"

        return dfw_wgs

    def make_basket(
        self,
        weight_meth: str = "equal",
        weights: List[float] = None,
        lback_meth: str = "xma",
        lback_periods: int = 21,
        ewgt: str = None,
        max_weight: float = 1.0,
        remove_zeros: bool = True,
        basket_name: str = "GLB_ALL",
    ):
        """
        Calculates all basket performance categories.

        Parameters
        ----------
        weight_meth : str
            method used for weighting constituent returns and carry. The parameter can
            receive either a single weight method or multiple weighting methods. The options
            are as follows:

            - [1] "equal": all constituents with non-NA returns have the same
            weight. This is the default.

            - [2] "fixed": weights are proportionate to a single
            list of values provided which are passed to argument `weights` (each value
            corresponds to a single contract).

            - [3] "invsd": weights based on inverse to
            standard deviations of recent returns.

            - [4] "values": weights proportionate to a
            panel of values of exogenous weight category.

            - [5] "inv_values": weights are
            inversely proportionate to of values of exogenous weight category.

        weights : List[float]
            single list of weights corresponding to the base tickers in `contracts`
            argument. This is only relevant for the fixed weight method.
        lback_meth : str
            look-back method for "invsd" weighting method. Default is Exponential MA,
            "ema". The alternative is simple moving average, "ma".
        lback_periods : int
            look-back periods for "invsd" weighting method. Default is 21.  Half-time
            for "xma" and full lookback period for "ma".
        ewgt : str
            Exogenous weight postfix that defines the weight value panel. Only needed
            for the 'values' or 'inv_values' method.
        max_weight : float
            maximum weight of a single contract. Default is 1, i.e zero restrictions.
            The purpose of the restriction is to limit concentration within the basket.
        remove_zeros : bool
            removes the zeros. Default is set to True.
        basket_name : str
            name of basket base ticker (analogous to contract name) to be used for
            return and (possibly) carry are calculated. Default is "GLB_ALL".
        """

        assert isinstance(weight_meth, str), "`weight_meth` must be string"

        self.exo_w_postfix: str = ewgt
        dfw_wgs = self.make_weights(
            weight_meth=weight_meth,
            weights=weights,
            lback_meth=lback_meth,
            lback_periods=lback_periods,
            ewgt=ewgt,
            max_weight=max_weight,
            remove_zeros=remove_zeros,
        )
        select = ["ticker", "real_date", "value"]

        dfw_wgs_copy = self.column_manager(df_cat=self.dfw_ret, dfw_wgs=dfw_wgs)

        dfw_bret = self.dfw_ret.multiply(dfw_wgs_copy).sum(axis=1)
        basket_ret = basket_name + "_" + self.ret
        store = [dfw_bret.to_frame(basket_ret)]
        if not hasattr(self, "cry_flag"):
            raise ValueError("Please initialise the class to use this method. ")
        self.cry_flag: bool
        if self.cry_flag:
            cry_list = []
            for cr in self.cry:
                dfw_wgs_copy = self.column_manager(
                    df_cat=self.dfws_cry[cr], dfw_wgs=dfw_wgs
                )
                self.dfws_cry: Dict[str, pd.DataFrame]
                dfw_bcry: pd.Series = (
                    self.dfws_cry[cr].multiply(dfw_wgs_copy).sum(axis=1)
                )
                tkr = basket_name + "_" + cr
                cry_list.append(dfw_bcry.to_frame(tkr))

            store += cry_list

        df_retcry = pd.concat(store, axis=1)
        self.dict_retcry[basket_name] = df_retcry

        self.dict_wgs[basket_name] = self.column_weights(dfw_wgs)

    def weight_visualiser(
        self,
        basket_name,
        start_date: str = None,
        end_date: str = None,
        subplots: bool = True,
        facet_grid: bool = False,
        scatter: bool = False,
        all_tickers: bool = True,
        single_ticker: str = None,
        percentage_change: bool = False,
        size: Tuple[int, int] = (7, 7),
    ):
        """
        Method used to visualise the weights associated with each contract in the
        basket.

        Parameters
        ----------
        basket_name : str
            name of basket whose weights are visualized
        start_date : str
            start date of he visualisation period.
        end_date : str
            end date of the visualization period.
        subplots : bool
            contract weights are displayed on different plots (True) or on a single plot
            (False).
        facet_grid : bool
            parameter used to break up the plot into multiple cartesian coordinate
            systems. If the basket consists of a high number of contracts, using the Facet
            Grid is recommended.
        scatter : bool
            if the facet_grid parameter is set to True there are two options: i) scatter
            plot if there a lot of blacklist periods; ii) line plot for continuous series.
        all_tickers : bool
            if True (default) all weights are displayed. If set to False `single-ticker`
            must be specified.
        single_ticker : str
            individual ticker for further, more detailed, analysis.
        percentage_change : bool
            graphical display used to further assimilate the fluctuations in the
            contract's weight. The graphical display is limited to a single contract.
            Therefore, pass the ticker into the parameter "single_ticker".
        size : Tuple[int, int]
            size of the plot. Default is (7, 7).
        """

        date_conv = lambda d: pd.Timestamp(d).strftime("%Y-%m-%d %X")
        try:
            dfw_wgs = self.dict_wgs[basket_name]
        except KeyError as e:
            print(f"Basket not found - call make_basket() method first: {e}.")
        else:
            if isinstance(start_date, str) and isinstance(end_date, str):
                self.date_check(start_date)
                self.date_check(end_date)
                start_date = date_conv(start_date)
                end_date = date_conv(end_date)
            elif isinstance(start_date, str):
                self.date_check(start_date)
                start_date = date_conv(start_date)
                end_date = dfw_wgs.index[-1]
            elif isinstance(end_date, str):
                self.date_check(end_date)
                start_date = dfw_wgs.index[0]
                end_date = date_conv(end_date)
            else:
                start_date = dfw_wgs.index[0]
                end_date = dfw_wgs.index[-1]

            error_1 = f"{start_date} unavailable in weight dataframe."
            c = dfw_wgs.index
            assert start_date in c, error_1
            error_2 = f"{end_date} unavailable in weight dataframe."
            assert end_date in c, error_2

            dfw_wgs = dfw_wgs.truncate(before=start_date, after=end_date)
            if not all_tickers:
                error_3 = "The parameter, 'single_ticker', must be a <str>."
                assert isinstance(single_ticker, str), error_3
                error_4 = (
                    f"Ticker not present in the weight dataframe. Available "
                    f"tickers: {list(dfw_wgs.columns)}."
                )
                assert single_ticker in dfw_wgs.columns, error_4
                dfw_wgs = dfw_wgs[[single_ticker]]

            if facet_grid:
                dfw_wgs.columns.name = "ticker"
                df_stack = dfw_wgs.stack().to_frame("value").reset_index()
                df_stack = df_stack.sort_values(["ticker", "real_date"])
                no_contracts = dfw_wgs.shape[1]
                facet_cols = 4 if no_contracts >= 8 else 3
                sns.set(rc={"figure.figsize": size})
                fg = sns.FacetGrid(
                    df_stack, col="ticker", col_wrap=facet_cols, sharey=True
                )

                scatter_error = (
                    f"Boolean object expected - instead received " f"{type(scatter)}."
                )
                assert isinstance(scatter, bool), scatter_error
                if scatter:
                    fg.map_dataframe(sns.scatterplot, x="real_date", y="value")
                else:
                    # Seaborn will linearly interpolate NaN values which is visually
                    # misleading. Therefore, aim to negate the operation.
                    fg.map_dataframe(
                        sns.lineplot,
                        x="real_date",
                        y="value",
                        hue=df_stack["value"].isna().cumsum(),
                        palette=["blue"] * df_stack["value"].isna().cumsum().nunique(),
                        estimator=None,
                        markers=True,
                    )

                equal_value = 1 / no_contracts
                fg.map(plt.axhline, y=equal_value, linestyle="--", color="gray", lw=0.5)
                # Set axes labels of individual charts.
                fg.set_axis_labels("", "")
                fg.set_titles(col_template="{col_name}")
                fg.fig.suptitle("Contract weights in basket", y=1.02)
            else:
                plt.rcParams["figure.figsize"] = size
                dfw_wgs.plot(
                    subplots=subplots, title="Weight Values Timestamp", legend=True
                )
                plt.xlabel("real_date, years")

            date_func = lambda d: pd.Timestamp(d).strftime("%Y-%m-%d")
            if percentage_change:
                error_5 = (
                    "Percentage change display is applied to a single ticker. Set "
                    "the parameter 'all_tickers' to False."
                )
                assert dfw_wgs.shape[1] == 1, error_5

                plt.rcParams["figure.figsize"] = size
                fig, ax = plt.subplots()
                dfw_pct = dfw_wgs.pct_change(periods=1) * 100
                n_index = np.array(list(map(date_func, dfw_pct.index)))
                dfw_pct = dfw_pct.set_index(keys=n_index)
                dfw_pct.plot(kind="bar", color="coral", ax=ax)
                ax.xaxis.set_major_locator(mdates.MonthLocator())
                plt.xticks(rotation=0)
                ax.set_ylabel("Percentage Change in weight.")
                ax.legend()

            plt.show()

    @staticmethod
    def column_split(df: pd.DataFrame):
        """
        Receives a dataframe with the columns 'ticker', 'real_date' and 'value' and
        returns a standardised dataframe with the columns 'cid', 'xcat', 'real_date' and
        'value. The 'ticker' column is broken up to produce the two new columns.

        Parameters
        ----------
        df : ~pandas.DataFrame

        Returns
        -------
        ~pandas.DataFrame
            standardised dataframe.
        """

        select = ["cid", "xcat", "real_date", "value"]

        cid_func = lambda t: t.split("_")[0]
        xcat_func = lambda t: "_".join(t.split("_")[1:])

        cids_w_df = list(map(cid_func, df["ticker"]))
        df["cid"] = np.array(cids_w_df)

        df = df.rename(columns={"ticker": "xcat"})
        df["xcat"] = np.array(list(map(xcat_func, df["xcat"])))

        df = df[select]
        return df

    def return_basket(self, basket_names: Union[str, List[str]] = None):
        """
        Return standardized dataframe of basket performance data based on one or more
        weighting methods.

        Parameters
        ----------
        basket_names : str or List[str]
            single basket name or list for which performance data are to be returned. If
            none is given all baskets added to the instance are selected.

        Returns
        -------
        QuantamentalDataFrame
            standardized DataFrame with the basket return and (possibly) carry data in
            standard form, i.e. columns 'cid', 'xcat', 'real_date' and 'value'.
        """

        if basket_names is None:
            basket_names = list(self.dict_retcry.keys())

        basket_error = "String or List of basket names expected."
        assert isinstance(basket_error, (list, str)), basket_error
        if isinstance(basket_names, str):
            basket_names = [basket_names]

        ret_baskets = []
        for b in basket_names:
            try:
                dfw_retcry = self.dict_retcry[b]
            except KeyError as e:
                print(f"Basket not found - call make_basket() method first: {e}.")
            else:
                qdf = QuantamentalDataFrame.from_wide(dfw_retcry)
                ret_baskets.append(qdf)

        return_df = QuantamentalDataFrame.from_qdf_list(ret_baskets)
        return QuantamentalDataFrame(
            df=return_df, _initialized_as_categorical=self._as_categorical
        ).to_original_dtypes()

    def return_weights(self, basket_names: Union[str, List[str]] = None):
        """
        Return the standardised dataframe containing the corresponding weights used to
        compute the basket.

        Parameters
        ----------
        basket_names : str or List[str]
            single basket name or list for which performance data are to be returned. If
            none is given all baskets added to the instance are selected.

        Returns
        -------
        QuantamentalDataFrame
            standardized DataFrame with basket weights.
        """

        if basket_names is None:
            basket_names = list(self.dict_wgs.keys())

        basket_error = "String or List of basket names expected."
        assert isinstance(basket_error, (list, str)), basket_error
        if isinstance(basket_names, str):
            basket_names = [basket_names]

        weight_baskets = []
        select = ["cid", "xcat", "real_date", "value"]

        for b in basket_names:
            try:
                dfw_wgs: pd.DataFrame = self.dict_wgs[b]
            except KeyError as e:
                print(f"Basket not found - call make_basket() method first: {e}.")
            else:
                pfx = f"_{b}_WGT"
                qdf = QuantamentalDataFrame.from_wide(dfw_wgs).rename_xcats(postfix=pfx)
                weight_baskets.append(qdf)

        return QuantamentalDataFrame(
            df=QuantamentalDataFrame.from_qdf_list(weight_baskets),
            _initialized_as_categorical=self._as_categorical,
        ).to_original_dtypes()


if __name__ == "__main__":
    cids = ["AUD", "GBP", "NZD", "USD"]
    xcats = [
        "FXXR_NSA",
        "FXCRY_NSA",
        "FXCRR_NSA",
        "EQXR_NSA",
        "EQCRY_NSA",
        "EQCRR_NSA",
        "FXWBASE_NSA",
        "EQWBASE_NSA",
    ]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    df_cids.loc["AUD"] = ["2000-01-01", "2022-03-14", 0, 1]
    df_cids.loc["GBP"] = ["2001-01-01", "2022-03-14", 0, 2]
    df_cids.loc["NZD"] = ["2002-01-01", "2022-03-14", 0, 3]
    df_cids.loc["USD"] = ["2000-01-01", "2022-03-14", 0, 4]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["FXXR_NSA"] = ["2010-01-01", "2022-03-14", 0, 1, 0, 0.2]
    df_xcats.loc["FXCRY_NSA"] = ["2010-01-01", "2022-03-14", 1, 1, 0.9, 0.2]
    df_xcats.loc["FXCRR_NSA"] = ["2010-01-01", "2022-03-14", 0.5, 0.8, 0.9, 0.2]
    df_xcats.loc["EQXR_NSA"] = ["2010-01-01", "2022-03-14", 0.5, 2, 0, 0.2]
    df_xcats.loc["EQCRY_NSA"] = ["2010-01-01", "2022-03-14", 2, 1.5, 0.9, 0.5]
    df_xcats.loc["EQCRR_NSA"] = ["2010-01-01", "2022-03-14", 1.5, 1.5, 0.9, 0.5]
    df_xcats.loc["FXWBASE_NSA"] = ["2010-01-01", "2022-02-01", 1, 1.5, 0.8, 0.5]
    df_xcats.loc["EQWBASE_NSA"] = ["2010-01-01", "2022-02-01", 1, 1.5, 0.9, 0.5]

    random.seed(2)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    black = {"AUD": ["2010-01-01", "2013-12-31"], "GBP": ["2010-01-01", "2013-12-31"]}
    contracts = ["AUD_FX", "AUD_EQ", "NZD_FX", "GBP_EQ", "USD_EQ"]
    gdp_figures = [17.0, 17.0, 41.0, 9.0, 250.0]

    contracts_1 = ["AUD_FX", "GBP_FX", "NZD_FX", "USD_EQ"]

    # First test. Multiple carries. Equal weight method.
    # The main aspect to check in the code is that basket performance has been applied to
    # both the return category and the multiple carry categories.
    dfd["grading"] = np.ones(dfd.shape[0])

    basket_1 = Basket(
        df=dfd,
        contracts=contracts_1,
        ret="XR_NSA",
        cry=["CRY_NSA", "CRR_NSA"],
        blacklist=black,
    )
    basket_1.make_basket(weight_meth="equal", max_weight=0.55, basket_name="GLB_EQUAL")

    basket_1.make_basket(
        weight_meth="fixed",
        max_weight=0.55,
        weights=[1 / 6, 1 / 6, 1 / 6, 1 / 2],
        basket_name="GLB_FIXED",
    )

    # show the weights of the GLB_FIXED basket

    basket_1.weight_visualiser(basket_name="GLB_FIXED", subplots=False, size=(10, 5))



from .metrics import (
    regression_accuracy,
    regression_balanced_accuracy,
    panel_significance_probability,
    sharpe_ratio,
    sortino_ratio,
    correlation_coefficient,
    create_panel_metric,
)

from .scorers import (
    neg_mean_abs_corr
)

__all__ = [
    "regression_accuracy",
    "regression_balanced_accuracy",
    "panel_significance_probability",
    "neg_mean_abs_corr",
    "sharpe_ratio",
    "sortino_ratio",
    "correlation_coefficient",
    "create_panel_metric",
]


"""
Function for visualising the `eop_lag`, `mop_lag` or `grading` metrics for a given set
of cross sections and extended categories.
"""

import pandas as pd

from typing import List, Tuple, Optional
from macrosynergy.management.simulate import make_test_df

import macrosynergy.visuals as msv


def view_metrics(
    df: pd.DataFrame,
    xcat: str,
    cids: Optional[List[str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    freq: str = "M",
    agg: str = "mean",
    metric: str = "eop_lag",
    title: Optional[str] = None,
    figsize: Optional[Tuple[float]] = (14, None),
) -> None:
    """
    A function to visualise the `eop_lag`, `mop_lag` or `grading` metrics for a given
    JPMaQS dataset. It generates a heatmap, where the x-axis is the observation date,
    the y-axis is the ticker, and the colour is the lag value.

    Parameters
    ----------
    df : ~pandas.Dataframe
        standardized DataFrame with the necessary columns: 'cid', 'xcat', 'real_date'
        and 'grading', 'eop_lag' or 'mop_lag'.
    xcat : str
        extended category whose lags are to be visualized.
    cids : List[str]
        cross sections to visualize. Default is all in DataFrame.
    start : str
        earliest date in ISO format. Default is earliest available.
    end : str
        latest date in ISO format. Default is latest available.
    metric : str
        name of metric to be visualized. Must be "eop_lag" (default) "mop_lag" or
        "grading".
    freq : str
        frequency of data. Must be one of "D", "W", "M", "Q", "A". Default is "M".
    agg : str
        aggregation method. Must be one of "mean" (default), "median", "min", "max",
        "first" or "last".
    title : str
        string of chart title; if none given default title is printed.
    figsize : Tuple[float]
        Tuple (w, h) of width and height of graph. Default is None, meaning it is set in
        accordance with df.

    Raises
    ------
    TypeError
        if any of the inputs are of the wrong type.
    ValueError
        if any of the inputs are semantically incorrect.

    Returns
    -------
    None
        None
    """

    msv.view_metrics(
        df=df,
        xcat=xcat,
        cids=cids,
        start=start,
        end=end,
        freq=freq,
        agg=agg,
        metric=metric,
        title=title,
        figsize=figsize,
    )


if __name__ == "__main__":
    test_cids: List[str] = ["USD", "EUR", "GBP"]
    test_xcats: List[str] = ["FX", "IR"]
    dfE: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="sharp-hill"
    )

    dfM: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="four-bit-sine"
    )

    dfG: pd.DataFrame = make_test_df(cids=test_cids, xcats=test_xcats, style="sine")

    dfE.rename(columns={"value": "eop_lag"}, inplace=True)
    dfM.rename(columns={"value": "mop_lag"}, inplace=True)
    dfG.rename(columns={"value": "grading"}, inplace=True)
    mergeon = ["cid", "xcat", "real_date"]
    dfx: pd.DataFrame = pd.merge(pd.merge(dfE, dfM, on=mergeon), dfG, on=mergeon)

    view_metrics(
        df=dfx,
        xcat="FX",
    )
    view_metrics(
        df=dfx,
        xcat="IR",
        metric="mop_lag",
    )
    view_metrics(
        df=dfx,
        xcat="IR",
        metric="grading",
    )



import numpy as np
import pandas as pd

from macrosynergy.learning.forecasting.linear_model.lad_regressors import LADRegressor
from macrosynergy.learning.forecasting.weighted_regressors import (
    SignWeightedRegressor,
    TimeWeightedRegressor,
)


class SignWeightedLADRegressor(SignWeightedRegressor):
    def __init__(
        self,
        fit_intercept: bool = True,
        positive=False,
        alpha=0,
        shrinkage_type="l1",
        tol=None,
        maxiter=None,
    ):
        """
        LAD regressor with sign-weighted loss.

        Parameters
        ----------
        fit_intercept : bool, default=True
            Whether or not to add an intercept to the model.
        positive: bool, default=False
            Whether or not to enforce non-negativity of model weights,
            with exception to a model intercept.
        alpha: float, default=0
            Shrinkage hyperparameter.
        shrinkage_type: str, default="l1"
            Type of shrinkage regularization to perform.
        tol: float, default=None
            Tolerance for convergence of the learning algorithm (SLSQP).
            This is passed into the 'tol' parameter of the scipy.optimize.minimize
            function.
        maxiter: int, default=None
            Maximum number of iterations for the learning algorithm (SLSQP).
            This is passed into the 'maxiter' key of the options dictionary in the
            scipy.optimize.minimize function.


        Notes
        -----
        A dependent variable is modelled as a linear combination of the input features.
        The weights associated with each feature (and the intercept) are determined by
        finding the weights that minimise the weighted average absolute model residuals,
        where the weighted average is based on inverse frequency of the sign of the
        dependent variable.

        By weighting the contribution of different training samples based on the
        sign of the label, the model is encouraged to learn equally from both positive and
        negative return samples, irrespective of class imbalance. If there are more
        positive targets than negative targets in the training set, then the negative
        target samples are given a higher weight in the model training process.
        The opposite is true if there are more negative targets than positive targets.
        """
        super().__init__(
            model=LADRegressor(
                fit_intercept=fit_intercept,
                positive=positive,
                alpha=alpha,
                shrinkage_type=shrinkage_type,
                tol=tol,
                maxiter=maxiter,
            ),
        )
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.alpha = alpha
        self.shrinkage_type = shrinkage_type
        self.tol = tol
        self.maxiter = maxiter

    def set_params(self, **params):
        """
        Setter method to update the parameters of the `SignWeightedLADRegressor`.

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The `SignWeightedLADRegressor` instance with updated parameters.
        """
        super().set_params(**params)

        relevant_params = {"fit_intercept", "positive", "alpha", "shrinkage_type"}

        if relevant_params.intersection(params):
            self.model = LADRegressor(
                fit_intercept=self.fit_intercept,
                positive=self.positive,
                alpha=self.alpha,
                shrinkage_type=self.shrinkage_type,
            )

        return self


class TimeWeightedLADRegressor(TimeWeightedRegressor):
    def __init__(
        self,
        fit_intercept=True,
        positive=False,
        half_life=21 * 12,
        alpha=0,
        shrinkage_type="l1",
        tol=None,
        maxiter=None,
    ):
        """
        LAD regressor with time-weighted loss.

        Parameters
        ----------
        fit_intercept : bool, default=True
            Whether or not to add an intercept to the model.
        positive: bool, default=False
            Whether or not to enforce non-negativity of model weights,
            with exception to a model intercept.
        half_life: int, default=21 * 12
            Half-life of the exponential decay function used to weight the
            contribution of different training samples based on the time of
            the sample.
        alpha: float, default=0
            Shrinkage hyperparameter.
        shrinkage_type: str, default="l1"
            Type of shrinkage regularization to perform.
        tol: float, default=None
            Tolerance for convergence of the learning algorithm (SLSQP).
            This is passed into the 'tol' parameter of the scipy.optimize.minimize
            function.
        maxiter: int, default=None
            Maximum number of iterations for the learning algorithm (SLSQP).
            This is passed into the 'maxiter' key of the options dictionary in the
            scipy.optimize.minimize function.

        Notes
        -----
        A dependent variable is modelled as a linear combination of the input features.
        The weights associated with each feature (and the intercept) are determined by
        finding the weights that minimise the weighted average absolute model residuals,
        where the weighted average is based on an exponential decay specified by the
        half-life parameter, where more recent samples are given higher weight.

        By weighting the contribution of different training samples based on the
        timestamp, the model is encouraged to prioritise more recent samples in the
        model training process. The half-life denotes the number of time periods in units
        of the native dataset frequency for the weight attributed to the most recent sample
        (one) to decay by half.
        """
        super().__init__(
            model=LADRegressor(
                fit_intercept=fit_intercept,
                positive=positive,
                alpha=alpha,
                shrinkage_type=shrinkage_type,
                tol=tol,
                maxiter=maxiter,
            ),
            half_life=half_life,
        )
        self.fit_intercept = fit_intercept
        self.positive = positive
        self.half_life = half_life
        self.alpha = alpha
        self.shrinkage_type = shrinkage_type
        self.tol = tol
        self.maxiter = maxiter

    def set_params(self, **params):
        """
        Setter method to update the parameters of the `TimeWeightedLADRegressor`.

        Parameters
        ----------
        **params : dict
            Dictionary of parameters to update.

        Returns
        -------
        self
            The `TimeWeightedLADRegressor` instance with updated parameters.
        """
        super().set_params(**params)

        relevant_params = {"fit_intercept", "positive", "alpha", "shrinkage_type"}

        if relevant_params.intersection(params):
            self.model = LADRegressor(
                fit_intercept=self.fit_intercept,
                positive=self.positive,
                alpha=self.alpha,
                shrinkage_type=self.shrinkage_type,
            )

        return self


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()

    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    # Fit SWLAD - no regularization
    model = SignWeightedLADRegressor(
        fit_intercept=True, positive=False, alpha=0, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")

    # Fit SWLAD - regularization
    model = SignWeightedLADRegressor(
        fit_intercept=True, positive=False, alpha=1, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")

    # Fit TWLAD - no regularization
    model = TimeWeightedLADRegressor(
        fit_intercept=True, positive=False, half_life=36, alpha=0, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")

    # Fit TWLAD - regularization
    model = TimeWeightedLADRegressor(
        fit_intercept=True, positive=False, half_life=36, alpha=1, shrinkage_type="l1"
    )
    model.fit(X_train, y_train)
    print(f"Intercept: {model.intercept_}, Coefficients: {model.coef_}")



from abc import ABC, abstractmethod

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from macrosynergy.compat import OneToOneFeatureMixin


class BasePanelScaler(BaseEstimator, TransformerMixin, OneToOneFeatureMixin, ABC):
    """
    Base class for scaling a panel of features in a learning pipeline.

    Parameters
    ----------
    type : str, default="panel"
        The panel dimension over which the scaling is applied. Options are
        "panel" and "cross_section".

    Notes
    -----
    Learning algorithms can benefit from scaling each feature to a similar range. This
    ensures they consider each feature equally in the model training process. It can
    also encourage faster convergence of an optimization algorithm.
    """

    def __init__(self, type="panel"):
        # Checks
        if not isinstance(type, str):
            raise TypeError("`type` must be a string.")
        if type not in ["panel", "cross_section"]:
            raise ValueError("`type` must be either 'panel' or 'cross_section'.")

        # Attributes
        self.type = type
        self.n_features_in_ = None
        self.feature_names_in_ = None

    def fit(self, X, y=None):
        """
        Fit method to learn training set quantities for feature scaling.

        Parameters
        ----------
        X : pd.DataFrame
            The feature matrix.
        y : pd.Series or pd.DataFrame, default=None
            The target vector.

        Returns
        -------
        self
            The fitted scaler.
        """
        # Checks
        self._check_fit_params(X, y)
        self.n_features_in_ = X.shape[1]
        self.feature_names_in_ = X.columns

        # Set up hash table for storing statistics
        unique_cross_sections = X.index.get_level_values(0).unique()
        self.statistics: dict = {
            cross_section: {feature_name: None for feature_name in X.columns}
            for cross_section in unique_cross_sections
        }
        self.statistics["panel"] = {feature_name: None for feature_name in X.columns}

        # Extract statistics for each feature
        for feature in X.columns:
            if self.type == "cross_section":
                # Get unique training cross-sections
                unique_cross_sections = X.index.get_level_values(0).unique()
                for cross_section in unique_cross_sections:
                    self.statistics[cross_section][feature] = self.extract_statistics(
                        X.loc[cross_section], feature
                    )
            self.statistics["panel"][feature] = self.extract_statistics(X, feature)

        return self

    def transform(self, X):
        """
        Transform method to scale the input data based on extracted training statistics.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.

        Returns
        -------
        X_transformed : pandas.DataFrame
            The feature matrix with scaled features.
        """
        # Checks
        self._check_transform_params(X)

        # Transform the data
        unique_cross_sections = X.index.get_level_values(0).unique()
        scaled_columns = []
        for feature in X.columns:
            if self.type == "cross_section":
                # Scale each cross-section based on stored cross sectional statistics in abstract method
                # If the cross-section is not in the statistics dictionary, use the panel statistics
                X_transformed = pd.concat(
                    [
                        self.scale(
                            X.loc[cross_section],
                            feature,
                            self.statistics.get(
                                cross_section, self.statistics["panel"]
                            )[feature],
                        )
                        for cross_section in unique_cross_sections
                    ],
                    axis=0,
                )
            else:
                # Scale the panel based on stored panel statistics in abstract method
                X_transformed = self.scale(
                    X, feature, self.statistics["panel"][feature]
                )
            # Add transformed column to list
            scaled_columns.append(X_transformed)

        # Concatenate the transformed columns
        X_transformed = pd.DataFrame(
            pd.concat(scaled_columns, axis=1).values, index=X.index, columns=X.columns
        )

        return X_transformed

    @abstractmethod
    def extract_statistics(self, X, feature):
        """
        Determine the relevant statistics for feature scaling.
        """
        pass

    @abstractmethod
    def scale(self, X, feature, statistics):
        """
        Scale the input data based on the relevant statistics.
        """
        pass

    def _check_fit_params(self, X, y):
        """
        Checks the input data for the fit method.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        y : pandas.Series or pandas.DataFrame
            The target vector.
        """
        # Checks only necessary on X
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the selector must be a pandas dataframe. ",
                "If used as part of an sklearn pipeline, ensure that previous steps ",
                "return a pandas dataframe.",
            )
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for PanelStandardScaler",
                " must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for PanelStandardScaler must not contain any "
                "missing values."
            )
        if not isinstance(X.index, pd.MultiIndex):
            raise ValueError("The input feature matrix for X must be multi-indexed.")
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")

    def _check_transform_params(self, X):
        """
        Input checks for the transform method.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError(
                "Input feature matrix for the scaler must be a pandas dataframe. "
                "If used as part of an sklearn pipeline, ensure that previous steps "
                "return a pandas dataframe."
            )
        if not X.apply(lambda x: pd.api.types.is_numeric_dtype(x)).all():
            raise ValueError(
                "All columns in the input feature matrix for PanelStandardScaler",
                " must be numeric.",
            )
        if X.isnull().values.any():
            raise ValueError(
                "The input feature matrix for PanelStandardScaler must not contain any "
                "missing values."
            )
        if not X.index.get_level_values(0).dtype == "object":
            raise TypeError("The outer index of X must be strings.")
        if not X.index.get_level_values(1).dtype == "datetime64[ns]":
            raise TypeError("The inner index of X must be datetime.date.")

        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "The input feature matrix must have the same number of columns as the "
                "training feature matrix."
            )
        if not X.columns.equals(self.feature_names_in_):
            raise ValueError(
                "The input feature matrix must have the same columns as the training "
                "feature matrix."
            )



from ..decorators import argvalidation


class NoneTypeMeta(type):
    """
    MetaClass to support type checks for `None`.
    """

    def __instancecheck__(cls, instance):
        return instance is None or isinstance(instance, type(None))


class NoneType(metaclass=NoneTypeMeta):
    """
    Custom class definition for a NoneType that supports type checks for `None`.
    """

    pass


class SubscriptableMeta(type):
    """
    Convenience metaclass to allow subscripting of methods on a class.
    """

    def __getitem__(cls, item):
        if hasattr(cls, item) and callable(getattr(cls, item)):
            return getattr(cls, item)
        else:
            raise KeyError(f"{item} is not a valid method name")


class ArgValidationMeta(type):
    def __new__(cls, name, bases, dct: dict):
        for key, value in dct.items():
            if callable(value):
                dct[key] = argvalidation(value)
        return super().__new__(cls, name, bases, dct)



import warnings
import datetime
from typing import (
    Dict,
    List,
    NamedTuple,
    Optional,
    Union,
)
import pandas as pd

from macrosynergy.management.utils import reduce_df, is_valid_iso_date
from macrosynergy.management.utils.df_utils import standardise_dataframe


class QDFArgs(NamedTuple):
    """
    Contains the QuantamentalDataFrame and associated arguments.
    """

    df: pd.DataFrame
    cids: List[str]
    xcats: List[str]
    metrics: List[str]
    start: str
    end: str


def validate_and_reduce_qdf(
    df: pd.DataFrame,
    cids: Optional[List[str]] = None,
    xcats: Optional[List[str]] = None,
    metrics: Optional[List[str]] = None,
    intersect: Optional[bool] = False,
    tickers: Optional[List[str]] = None,
    blacklist: Optional[Dict[str, List[str]]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
):
    """
    Validates the inputs to a function that takes a DataFrame as its first argument. The
    DataFrame is then reduced according to the inputs.

    Parameters
    ----------
    df : pd.DataFrame
        A DataFrame with the following columns: 'cid', 'xcat', 'real_date', and at least
        one metric from - 'value', 'grading', 'eop_lag', or 'mop_lag'.
    cids : List[str]
        A list of cids to select from the DataFrame. If None, all cids are selected.
    xcats : List[str]
        A list of xcats to select from the DataFrame. If None, all xcats are selected.
    metrics : List[str]
        A list of metrics to select from the DataFrame. If None, all metrics are
        selected.
    intersect : bool
        if True only retains cids that are available for all xcats. Default is False.
    tickers : List[str]
        A list of tickers that will be selected from the DataFrame if they exist,
        regardless of start, end, blacklist, and intersect arguments.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the data frame. If
        one cross-section has several blacklist periods append numbers to the cross-section
        code.
    start : str
        ISO-8601 formatted date string. Select data from this date onwards. If None, all
        dates are selected.
    end : str
        ISO-8601 formatted date string. Select data up to and including this date. If
        None, all dates are selected.

    Returns
    -------
    QDFArgs
        A NamedTuple that contains the validated arguments.
    """

    df: pd.DataFrame = df.copy()
    df = standardise_dataframe(df=df)

    metrics = _validate_metrics(df=df, metrics=metrics)

    cids_provided: bool = cids is not None
    xcats_provided: bool = xcats is not None
    missing_cids: List[str]
    missing_xcats: List[str]

    cids, missing_cids = _set_or_find_missing_in_df(
        df=df, col_name="cid", values=cids, param_name="cids"
    )
    xcats, missing_xcats = _set_or_find_missing_in_df(
        df=df, col_name="xcat", values=xcats, param_name="xcats"
    )

    start, end = _validate_start_and_end_dates(df, start, end)

    ticker_df: pd.DataFrame = pd.DataFrame()
    if tickers is not None:
        ticker_df = _get_ticker_df(df=df, tickers=tickers, metrics=metrics)

    df: pd.DataFrame
    r_xcats: List[str]
    r_cids: List[str]
    df, r_xcats, r_cids = reduce_df(
        df=df,
        cids=cids if isinstance(cids, list) else [cids],
        xcats=xcats if isinstance(xcats, list) else [xcats],
        intersect=intersect,
        start=start,
        end=end,
        blacklist=blacklist,
        out_all=True,
    )

    df: pd.DataFrame = pd.concat([df, ticker_df], axis=0)
    df = df.drop_duplicates()

    if (
        ((len(r_xcats) != len(xcats) - len(missing_xcats)) and xcats_provided)
        or ((len(r_cids) != len(cids) - len(missing_cids)) and cids_provided)
    ) and not intersect:
        m_cids: List[str] = list(set(cids).difference(set(r_cids), set(missing_cids)))
        m_xcats: List[str] = list(
            set(xcats).difference(set(r_xcats), set(missing_xcats))
        )

        warnings.warn(
            "The provided arguments resulted in a DataFrame that does not "
            "contain all the requested cids and xcats. "
            + (f"Missing cids: {m_cids}. " if m_cids else "")
            + (f"Missing xcats: {m_xcats}. " if m_xcats else "")
        )
        for m_cid in m_cids:
            cids.remove(m_cid)
        for m_xcat in m_xcats:
            xcats.remove(m_xcat)

    if df.empty:
        raise ValueError(
            "The arguments provided resulted in an "
            "empty DataFrame when filtered (see `reduce_df`)."
        )

    return QDFArgs(df, cids, xcats, metrics, start, end)


def _get_ticker_df(df: pd.DataFrame, tickers: List[str], metrics: Optional[List[str]]):
    """
    Filters a QuantamentalDataFrame by tickers.

    Parameters
    ----------
    df : pd.DataFrame
        a Pandas DataFrame.
    metrics : List[str]
        a list of metrics.
    tickers : List[str]
        a list of tickers.

    Returns
    -------
    pd.DataFrame
        the filtered DataFrame.
    """

    df_tickers: List[pd.DataFrame] = [pd.DataFrame()]
    for ticker in tickers:
        _cid, _xcat = ticker.split("_", 1)
        df_tickers.append(
            df.loc[
                (df["cid"] == _cid) & (df["xcat"] == _xcat),
                ["real_date", "cid", "xcat"] + metrics,
            ]
        )
    ticker_df: pd.DataFrame = pd.concat(df_tickers, axis=0)
    return ticker_df


def _validate_start_and_end_dates(df: pd.DataFrame, start: str, end: str):
    """
    Determines start and end dates for a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame to be filtered.
    start : str
        ISO-8601 formatted date string. If None, the earliest date in the DataFrame is
        used.
    end : str
        ISO-8601 formatted date string. If None, the latest date in the DataFrame is
        used.

    Returns
    -------
    Tuple[str]
        Tuple of start and end dates.
    """

    if start is None:
        start: str = pd.Timestamp(df["real_date"].min()).strftime("%Y-%m-%d")
    if end is None:
        end: str = pd.Timestamp(df["real_date"].max()).strftime("%Y-%m-%d")
    for var, name in [(start, "start"), (end, "end")]:
        if not is_valid_iso_date(var):
            raise ValueError(f"`{name}` must be a valid ISO date string")

    return start, end


def _validate_metrics(df: pd.DataFrame, metrics: List[str]):
    """
    Validates the metrics passed to a function.

    Parameters
    ----------
    df : pd.DataFrame
        a Pandas DataFrame.
    metrics : List[str]
        a list of metrics to be checked.

    Returns
    -------
    List[str]
        a list of metrics to be used.
    """

    required_columns: List[str] = ["real_date", "cid", "xcat"]
    if metrics is None:
        metrics: List[str] = list(set(df.columns) - set(required_columns))
    required_columns += metrics
    if not set(required_columns).issubset(set(df.columns)):
        raise ValueError(
            f"DataFrame must contain the following columns: {required_columns}"
        )
    return metrics


def _set_or_find_missing_in_df(
    df: pd.DataFrame, col_name: str, values: Optional[List], param_name: str
):
    """
    Returns the values passed to a function and a list of values that are not found in a
    specific column of a DataFrame. If values is None, all unique values in the
    DataFrame are returned.

    Parameters
    ----------
    df : pd.DataFrame
        a Pandas DataFrame.
    col_name : str
        name of column in the DataFrame.
    values : List
        list of values to be checked.
    param_name : str
        name of parameter passed to function.

    Returns
    -------
    List
        list of values that are not in the DataFrame.
    """

    missing_values: List = []
    if values is None:
        values = df[col_name].unique().tolist()
    else:
        missing_values = _find_missing_in_df(
            df=df, col_name=col_name, values=values, param_name=param_name
        )
    return values, missing_values


def _find_missing_in_df(df: pd.DataFrame, col_name: str, values: List, param_name: str):
    """
    Finds values in a list that are not in a specific column of a DataFrame.

    Parameters
    ----------
    df : pd.DataFrame
        a Pandas DataFrame.
    col_name : str
        name of column in the DataFrame.
    values : List
        list of values to be checked.
    param_name : str
        name of parameter passed to function.

    Returns
    -------
    List
        list of values that are not in the DataFrame.
    """

    missing: List = []
    if not set(values).issubset(set(df[col_name].unique())):
        # warn
        warnings.warn(
            f"The following {col_name}(s), passed in `{param_name}`,"
            " are not in the DataFrame `df`: "
            f"{list(set(values) - set(df[col_name].unique()))}."
        )
        missing = list(set(values) - set(df[col_name].unique()))

    return missing


from . import PYTHON_3_8_OR_LATER
import pandas as pd
from packaging import version

if PYTHON_3_8_OR_LATER:
    RESAMPLE_NUMERIC_ONLY = {"numeric_only": True}
    JOBLIB_RETURN_AS = {"return_as": "generator"}
    from sklearn.base import OneToOneFeatureMixin
else:
    RESAMPLE_NUMERIC_ONLY = {}
    JOBLIB_RETURN_AS = {}
    from sklearn.base import _OneToOneFeatureMixin as OneToOneFeatureMixin


if version.parse(pd.__version__) > version.parse("2.1.0"):
    PD_FUTURE_STACK = {"future_stack": True}
else:
    PD_FUTURE_STACK = {"dropna": False}

PD_NEW_DATE_FREQ: bool = version.parse(pd.__version__) > version.parse("2.1.4")

PD_OLD_RESAMPLE: bool = version.parse(pd.__version__) < version.parse("1.5.0")

PD_2_0_OR_LATER: bool = version.parse(pd.__version__) >= version.parse("2.0.0")



"""
Scikit-learn compatible performance metrics for model evaluation.
"""

import inspect
import numpy as np
import pandas as pd
import scipy.stats as stats

from macrosynergy.learning.random_effects import RandomEffects

from sklearn.metrics import accuracy_score, balanced_accuracy_score


def create_panel_metric(
    y_true,
    y_pred,
    sklearn_metric,
    type="panel",
):
    """
    Evaluation with a scikit-learn metric, respecting the panel structure.

    Parameters
    ----------
    y_true : pd.Series of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    sklearn_metric : callable
        A scikit-learn metric function. This function must accept two arguments, `y_true`
        and `y_pred`, which have the same meaning as the arguments passed to this function.
    type : str, default="panel"
        The panel dimension over which to compute the metric. Options are "panel",
        "cross_section" and "time_periods".

    Returns
    -------
    metric : float
        The computed metric.

    Notes
    -----
    This function is a wrapper around a scikit-learn metric, allowing it to be
    evaluated over different panel axes. For instance, the :math:`R^2` metric can be
    evaluated over the whole panel, across cross-sections or across time periods. Instead
    of re-implementing every scikit-learn metric so that evaluation over panel axes is
    possible, the `create_panel_metric` function allows any scikit-learn metric to be
    evaluated over different panel axes.
    """
    # Checks
    _check_metric_params(y_true, y_pred, type)

    if not callable(sklearn_metric):
        raise TypeError("sklearn_metric must be a callable")
    # check sklearn_metric has y_true and y_pred as arguments
    metric_args = inspect.signature(sklearn_metric).parameters
    if not all(arg in metric_args for arg in ["y_true", "y_pred"]):
        raise ValueError("sklearn_metric must accept y_true and y_pred as arguments")

    if type == "panel":
        return sklearn_metric(y_true, y_pred)
    elif type == "cross_section":
        metrics = []
        unique_cross_sections = y_true.index.get_level_values(0).unique()
        for cross_section in unique_cross_sections:
            cross_section_mask = y_true.index.get_level_values(0) == cross_section
            metrics.append(
                sklearn_metric(
                    y_true.values[cross_section_mask], y_pred[cross_section_mask]
                )
            )
        return np.mean(metrics)
    elif type == "time_periods":
        metrics = []
        unique_time_periods = y_true.index.get_level_values(1).unique()
        for time_period in unique_time_periods:
            time_period_mask = y_true.index.get_level_values(1) == time_period
            metrics.append(
                sklearn_metric(
                    y_true.values[time_period_mask], y_pred[time_period_mask]
                )
            )
        return np.mean(metrics)


def regression_accuracy(
    y_true,
    y_pred,
    type="panel",
):
    """
    Accuracy of signs between regression labels and predictions.

    Parameters
    ----------

    y_true : pd.Series of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    type : str, default="panel"
        The panel dimension over which to compute the accuracy. Options are "panel",
        "cross_section" and "time_periods".

    Returns
    -------

    accuracy : float
        The accuracy betweens signs of prediction-target pairs.

    Notes
    -----
    Accuracy can be calculated over the whole panel, considering all samples irrespective
    of cross-section or time period. It can be beneficial, however, to estimate the
    expected accuracy for a cross-section or time period instead.

    When type = "cross_section", the returned accuracy is the mean accuracy across
    cross-sections, an empirical estimate of the expected accuracy for a cross-section
    of interest.

    When type = "time_periods", the returned accuracy is the mean accuracy across time
    periods, an empirical estimate of the expected accuracy for a time period of interest.
    """
    # Checks
    _check_metric_params(y_true, y_pred, type)

    # Compute accuracy
    if type == "panel":
        return accuracy_score(y_true < 0, y_pred < 0)

    elif type == "cross_section":
        accuracies = []
        unique_cross_sections = y_true.index.get_level_values(0).unique()
        for cross_section in unique_cross_sections:
            cross_section_mask = y_true.index.get_level_values(0) == cross_section
            accuracies.append(
                accuracy_score(
                    y_true.values[cross_section_mask] < 0,
                    y_pred[cross_section_mask] < 0,
                )
            )
        return np.mean(accuracies)

    elif type == "time_periods":
        accuracies = []
        unique_time_periods = y_true.index.get_level_values(1).unique()
        for time_period in unique_time_periods:
            time_period_mask = y_true.index.get_level_values(1) == time_period
            accuracies.append(
                accuracy_score(
                    y_true.values[time_period_mask] < 0, y_pred[time_period_mask] < 0
                )
            )
        return np.mean(accuracies)


def regression_balanced_accuracy(
    y_true,
    y_pred,
    type="panel",
):
    """
    Balanced accuracy of signs between regression labels and predictions.

    Parameters
    ----------

    y_true : pd.Series of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    type : str, default="panel"
        The panel dimension over which to compute the balanced accuracy. Options are
        "panel", "cross_section" and "time_periods".

    Returns
    -------

    balanced_accuracy : float
        The balanced accuracy betweens signs of prediction-target pairs.

    Notes
    -----
    Balanced accuracy can be calculated over the whole panel, considering all samples
    irrespective of cross-section or time period. It can be beneficial, however, to
    estimate expected balanced accuracy for a cross-section or time period instead.

    When type = "cross_section", the returned balanced accuracy score is
    the mean balanced accuracy across cross-sections, an empirical estimate of the expected
    balanced accuracy for a cross-section of interest.

    When type = "time_periods", the returned balanced accuracy score is the mean balanced
    accuracy across time periods, an empirical estimate of the expected
    balanced accuracy for a time period of interest.
    """
    # Checks
    _check_metric_params(y_true, y_pred, type)

    # Compute balanced accuracy
    if type == "panel":
        return balanced_accuracy_score(y_true < 0, y_pred < 0)

    elif type == "cross_section":
        balanced_accuracies = []
        unique_cross_sections = y_true.index.get_level_values(0).unique()
        for cross_section in unique_cross_sections:
            cross_section_mask = y_true.index.get_level_values(0) == cross_section
            balanced_accuracies.append(
                balanced_accuracy_score(
                    y_true.values[cross_section_mask] < 0,
                    y_pred[cross_section_mask] < 0,
                )
            )
        return np.mean(balanced_accuracies)

    elif type == "time_periods":
        balanced_accuracies = []
        unique_time_periods = y_true.index.get_level_values(1).unique()
        for time_period in unique_time_periods:
            time_period_mask = y_true.index.get_level_values(1) == time_period
            balanced_accuracies.append(
                balanced_accuracy_score(
                    y_true.values[time_period_mask] < 0, y_pred[time_period_mask] < 0
                )
            )
        return np.mean(balanced_accuracies)


def panel_significance_probability(
    y_true,
    y_pred,
):
    """
    :math:`1 - pval` using the Macrosynergy panel (MAP) test for significance of
    correlation accounting for cross-sectional correlations.

    Parameters
    ----------

    y_true : pd.Series of shape (n_samples,)
        True regression labels.

    y_pred : pd.Series of shape (n_samples,)
        Predicted regression labels.

    Returns
    -------

    prob_significance : float
        The probability of significance of the relation between predictions and targets.

    Notes
    -----

    The (Ma)crosynergy (p)anel (MAP) test is a hypothesis test for the significance of
    a relation between two variables accounting for cross-sectional correlations. A
    period-specific random effects model is estimated, with a Wald test performed on the
    concerned coefficient. Since the test requires a panel structure, the inputs are
    required to be pd.Series, multi-indexed by cross-section and real date.
    """
    # Checks
    # y_true
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pd.Series")
    if not y_true.index.nlevels == 2:
        raise ValueError("y_true must be multi-indexed with two levels")
    if not y_true.index.get_level_values(0).dtype == "object":
        raise ValueError("y_true outer index must be strings")
    if not y_true.index.get_level_values(1).dtype == "datetime64[ns]":
        raise ValueError("y_true inner index must be datetime")

    # y_pred
    if not isinstance(y_pred, (pd.Series, np.ndarray)):
        raise TypeError("y_pred must be either pd.Series or np.ndarray")
    if isinstance(y_pred, np.ndarray):
        if y_pred.ndim != 1:
            raise ValueError(
                "y_pred must be 1-dimensional for 'panel_significance_probability'"
            )
        y_pred = pd.Series(y_pred, index=y_true.index)
    if not y_pred.index.nlevels == 2:
        raise ValueError("y_pred must be multi-indexed with two levels")
    if not y_pred.index.get_level_values(0).dtype == "object":
        raise ValueError("y_pred outer index must be strings")
    if not y_pred.index.get_level_values(1).dtype == "datetime64[ns]":
        raise ValueError("y_pred inner index must be datetime")
    if len(y_true) != len(y_pred):
        raise ValueError("y_true and y_pred must have the same length")
    if not y_true.index.equals(y_pred.index):
        raise ValueError("y_true and y_pred must have the same index")
    if y_pred.isnull().values.any():
        return 0

    re = RandomEffects(fit_intercept=False).fit(y_true, y_pred)
    pval = re.pvals.to_numpy().item()

    return 1 - pval


def sharpe_ratio(
    y_true,
    y_pred,
    binary=True,
    thresh=None,
    type="panel",
):
    """
    Sharpe ratio of a strategy where the trader goes long by a single unit when the
    predictions are positive and short by a single unit when the predictions are negative.

    Parameters
    ----------
    y_true : pd.Series of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    binary : bool, default=True
        Whether to consider only directional returns. If True, the portfolio returns only
        consider the sign of the predictions. If False, naive portfolio weights are
        determined. See Notes for more information on their calculation.
    thresh : float, default=None
        The threshold for portfolio weights in the case where binary = False.
    type : str, default="panel"
        The panel dimension over which to compute the Sharpe ratio. Options are "panel",
        "cross_section" and "time_periods".

    Returns
    -------
    sharpe_ratio : float
        The Sharpe ratio of the strategy.

    Notes
    -----
    A Sharpe ratio can be calculated over the whole panel, considering the mean and
    standard deviation of the returns irrespective of cross-section or time period. It can
    be beneficial, however, to estimate the expected Sharpe for a cross-section or time
    period instead.

    When type = "cross_section", the returned Sharpe ratio is the mean Sharpe ratio across
    cross-sections, an empirical estimate of the expected Sharpe ratio for a cross-section
    of interest.

    When type = "time_periods", the returned Sharpe ratio is the mean Sharpe ratio across
    time periods, an empirical estimate of the expected Sharpe ratio for a time period of
    interest.

    This metric can calculate the Sharpe ratio of either binary or non-binary strategies.
    When binary = False, predictions are normalized by their standard deviation in each
    time period. If thresh is not None, the resulting weights are clipped to the range
    [-thresh, thresh]. The resulting portfolio returns are the product of these derived
    weights and the true returns.
    """
    # Checks
    _check_metric_params(y_true, y_pred, type)

    if not isinstance(binary, bool):
        raise TypeError("binary must be a boolean")

    if not isinstance(thresh, (int, float)) and thresh is not None:
        raise TypeError("thresh must be an integer or float")
    if thresh is not None and thresh <= 0:
        raise ValueError("thresh must be positive")

    # Compute Sharpe ratio
    if binary:
        portfolio_returns = np.where(y_pred > 0, y_true, -y_true)
    else:
        # Massage y_pred into portfolio weights
        if not isinstance(y_pred, pd.Series):
            y_pred = pd.Series(y_pred, index=y_true.index)
        if thresh is not None:
            portfolio_weights = (
                y_pred / y_pred.groupby(level=1).transform("std")
            ).clip(lower=-thresh, upper=thresh)
        else:
            portfolio_weights = y_pred / y_pred.groupby(level=1).transform("std")
        portfolio_returns = portfolio_weights * y_true

    if type == "panel":
        average_return = np.mean(portfolio_returns)
        std_return = np.std(portfolio_returns)

        if std_return == 0:
            std_return = 1

        sharpe_ratio = average_return / std_return

        return sharpe_ratio

    elif type == "cross_section":
        sharpe_ratios = []
        unique_cross_sections = y_true.index.get_level_values(0).unique()
        for cross_section in unique_cross_sections:
            cross_section_mask = y_true.index.get_level_values(0) == cross_section
            cross_section_returns = portfolio_returns[cross_section_mask]
            average_return = np.mean(cross_section_returns)
            std_return = np.std(cross_section_returns)

            if std_return == 0:
                std_return = 1

            sharpe_ratio = average_return / std_return

            sharpe_ratios.append(sharpe_ratio)

        return np.mean(sharpe_ratios)

    elif type == "time_periods":
        sharpe_ratios = []
        unique_time_periods = y_true.index.get_level_values(1).unique()
        for time_period in unique_time_periods:
            time_period_mask = y_true.index.get_level_values(1) == time_period
            time_period_returns = portfolio_returns[time_period_mask]
            average_return = np.mean(time_period_returns)
            std_return = np.std(time_period_returns)

            if std_return == 0:
                std_return = 1

            sharpe_ratio = average_return / std_return

            sharpe_ratios.append(sharpe_ratio)

        return np.mean(sharpe_ratios)


def sortino_ratio(
    y_true,
    y_pred,
    binary=True,
    thresh=None,
    type="panel",
):
    """
    Sortino ratio of a strategy where the trader goes long when the predictions are
    positive by a single unit and short by a single unit when the predictions are negative.

    Parameters
    ----------
    y_true : pd.Series of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    binary : bool, default=True
        Whether to consider only directional returns. If True, the portfolio returns only
        consider the sign of the predictions. If False, naive portfolio weights are
        determined. See Notes for more information on their calculation.
    thresh : float, default=None
        The threshold for portfolio weights in the case where binary = False.
    type : str, default="panel"
        The panel dimension over which to compute the Sharpe ratio. Options are "panel",
        "cross_section" and "time_periods".

    Returns
    -------
    sortino_ratio : float
        The Sortino ratio of the strategy.

    Notes
    -----
    A Sortino ratio can be calculated over the whole panel, considering the mean and
    downside standard deviation of the returns irrespective of cross-section or time
    period. It can be beneficial, however, to estimate the expected Sortino for a
    cross-section or time period instead.

    When type = "cross_section", the returned Sortino ratio is the mean Sortino ratio
    across cross-sections, an empirical estimate of the expected Sortino ratio for a
    cross-section of interest.

    When type = "time_periods", the returned Sortino ratio is the mean Sortino ratio
    across time periods, an empirical estimate of the expected Sortino ratio for a time
    period of interest.

    This metric can calculate the Sortino ratio of either binary or non-binary strategies.
    When binary = False, predictions are normalized by their standard deviation in each
    time period. If thresh is not None, the resulting weights are clipped to the range
    [-thresh, thresh]. The resulting portfolio returns are the product of these derived
    weights and the true returns.
    """
    # Checks
    _check_metric_params(y_true, y_pred, type)

    if not isinstance(binary, bool):
        raise TypeError("binary must be a boolean")

    if not isinstance(thresh, (int, float)) and thresh is not None:
        raise TypeError("thresh must be an integer or float")
    if thresh is not None and thresh <= 0:
        raise ValueError("thresh must be positive")

    # Compute Sortino ratio
    if binary:
        portfolio_returns = np.where(y_pred > 0, y_true, -y_true)
    else:
        # Massage y_pred into portfolio weights
        if not isinstance(y_pred, pd.Series):
            y_pred = pd.Series(y_pred, index=y_true.index)
        if thresh is not None:
            portfolio_weights = (
                y_pred / y_pred.groupby(level=1).transform("std")
            ).clip(lower=-thresh, upper=thresh)
        else:
            portfolio_weights = y_pred / y_pred.groupby(level=1).transform("std")
        portfolio_returns = portfolio_weights * y_true

    if type == "panel":
        negative_returns = portfolio_returns[portfolio_returns < 0]
        average_return = np.mean(portfolio_returns)
        denominator = np.std(negative_returns)

        if denominator == 0 or np.isnan(denominator):
            denominator = 1

        sortino_ratio = average_return / denominator

        return sortino_ratio

    elif type == "cross_section":
        sortino_ratios = []
        unique_cross_sections = y_true.index.get_level_values(0).unique()
        for cross_section in unique_cross_sections:
            cross_section_mask = y_true.index.get_level_values(0) == cross_section
            cross_section_returns = portfolio_returns[cross_section_mask]
            negative_returns = cross_section_returns[cross_section_returns < 0]
            average_return = np.mean(cross_section_returns)
            denominator = np.std(negative_returns)

            if denominator == 0 or np.isnan(denominator):
                denominator = 1

            sortino_ratio = average_return / denominator

            sortino_ratios.append(sortino_ratio)

        return np.mean(sortino_ratios)

    elif type == "time_periods":
        sortino_ratios = []
        unique_time_periods = y_true.index.get_level_values(1).unique()
        for time_period in unique_time_periods:
            time_period_mask = y_true.index.get_level_values(1) == time_period
            time_period_returns = portfolio_returns[time_period_mask]
            negative_returns = time_period_returns[time_period_returns < 0]
            average_return = np.mean(time_period_returns)
            denominator = np.std(negative_returns)

            if denominator == 0 or np.isnan(denominator):
                denominator = 1

            sortino_ratio = average_return / denominator

            sortino_ratios.append(sortino_ratio)

        return np.mean(sortino_ratios)


def correlation_coefficient(
    y_true,
    y_pred,
    correlation_type="pearson",
    type="panel",
):
    """
    Correlation coefficient between true and predicted regression labels.

    Parameters
    ----------
    y_true : array-like of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    correlation_type : str, default="pearson"
        The type of correlation coefficient to compute. Options are "pearson", "spearman"
        and "kendall".

    Returns
    -------
    correlation : float
        The correlation coefficient between true and predicted regression labels.

    Notes
    -----
    A correlation coefficient can be calculated over the whole panel, considering all
    samples irrespective of cross-section or time period. It can be beneficial, however,
    to estimate the expected correlation coefficient for a cross-section or time period
    instead.

    When type = "cross_section", the returned correlation coefficient is the mean
    correlation coefficient across cross-sections, an empirical estimate of the expected
    correlation coefficient for a cross-section of interest.

    When type = "time_periods", the returned correlation coefficient is the mean correlation
    coefficient across time periods, an empirical estimate of the expected correlation
    coefficient for a time period of interest.
    """
    # Checks
    _check_metric_params(y_true, y_pred, type)

    if not isinstance(correlation_type, str):
        raise TypeError("correlation_type must be a string")
    if correlation_type not in ["pearson", "spearman", "kendall"]:
        raise ValueError(
            "Invalid correlation type. Options are 'pearson', 'spearman' and 'kendall'."
        )

    # Compute correlation coefficient
    if type == "panel":
        if correlation_type == "pearson":
            correlation = stats.pearsonr(y_true, y_pred)[0]
        elif correlation_type == "spearman":
            correlation = stats.spearmanr(y_true, y_pred)[0]
        elif correlation_type == "kendall":
            correlation = stats.kendalltau(y_true, y_pred)[0]

        return correlation

    elif type == "cross_section":
        correlations = []
        unique_cross_sections = y_true.index.get_level_values(0).unique()
        for cross_section in unique_cross_sections:
            cross_section_mask = y_true.index.get_level_values(0) == cross_section
            if correlation_type == "pearson":
                correlation = stats.pearsonr(
                    y_true.values[cross_section_mask], y_pred[cross_section_mask]
                )[0]
            elif correlation_type == "spearman":
                correlation = stats.spearmanr(
                    y_true.values[cross_section_mask], y_pred[cross_section_mask]
                )[0]
            elif correlation_type == "kendall":
                correlation = stats.kendalltau(
                    y_true.values[cross_section_mask], y_pred[cross_section_mask]
                )[0]

            correlations.append(correlation)

        return np.mean(correlations)

    elif type == "time_periods":
        correlations = []
        unique_time_periods = y_true.index.get_level_values(1).unique()
        for time_period in unique_time_periods:
            time_period_mask = y_true.index.get_level_values(1) == time_period
            if correlation_type == "pearson":
                correlation = stats.pearsonr(
                    y_true.values[time_period_mask], y_pred[time_period_mask]
                )[0]
            elif correlation_type == "spearman":
                correlation = stats.spearmanr(
                    y_true.values[time_period_mask], y_pred[time_period_mask]
                )[0]
            elif correlation_type == "kendall":
                correlation = stats.kendalltau(
                    y_true.values[time_period_mask], y_pred[time_period_mask]
                )[0]

            correlations.append(correlation)

        return np.mean(correlations)


def _check_metric_params(
    y_true,
    y_pred,
    type,
):
    """
    Generic input validation for model evaluation metrics.

    Parameters
    ----------
    y_true : pd.Series of shape (n_samples,)
        True regression labels.
    y_pred : array-like of shape (n_samples,)
        Predicted regression labels.
    type : str
        The panel dimension over which to compute the metric. Options are "panel",
        "cross_section" and "time_periods".
    """
    # y_true
    if not isinstance(y_true, pd.Series):
        raise TypeError("y_true must be a pd.Series")
    if not y_true.index.nlevels == 2:
        raise ValueError("y_true must be multi-indexed with two levels")
    if not y_true.index.get_level_values(0).dtype == "object":
        raise ValueError("y_true outer index must be strings")
    if not y_true.index.get_level_values(1).dtype == "datetime64[ns]":
        raise ValueError("y_true inner index must be datetime")

    # y_pred
    if not isinstance(y_pred, (np.ndarray, pd.Series)):
        raise TypeError("y_pred must be either a numpy array or pandas series")
    if len(y_true) != len(y_pred):
        raise ValueError("y_true and y_pred must have the same length")

    # type
    if not isinstance(type, str):
        raise TypeError("type must be a string")
    if type not in ["panel", "cross_section", "time_periods"]:
        raise ValueError(
            "Invalid type. Options are 'panel', 'cross_section' and 'time_periods'"
        )


if __name__ == "__main__":
    import macrosynergy.management as msm
    from macrosynergy.management.simulate import make_qdf
    from sklearn.linear_model import LinearRegression

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "BMXR"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2012-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["BMXR"] = ["2012-01-01", "2020-12-31", 1, 2, 0.95, 1]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    Xy = msm.categories_df(
        df=dfd, xcats=xcats, cids=cids, freq="M", lag=1, xcat_aggs=["last", "sum"]
    ).dropna()
    X = Xy.iloc[:, :-1]
    y = Xy.iloc[:, -1]

    lr = LinearRegression()
    lr.fit(X, y)

    # Accuracies
    print(
        "\nAccuracy over panel: "
        f"{regression_accuracy(y, lr.predict(X), type='panel')}"
    )
    print(
        "Cross-sectional accuracy: "
        f"{regression_accuracy(y, lr.predict(X), type='cross_section')}"
    )
    print(
        "Periodic accuracy: "
        f"{regression_accuracy(y, lr.predict(X), type='time_periods')}"
    )

    # Un-annualized Sharpe ratios
    print(
        "\nUn-annualized Sharpe over panel: "
        f"{sharpe_ratio(y, lr.predict(X), type='panel')}"
    )
    print(
        "Cross-sectional un-annualized Sharpe: "
        f"{sharpe_ratio(y, lr.predict(X), type='cross_section')}"
    )
    print(
        "Periodic un-annualized Sharpe: "
        f"{sharpe_ratio(y, lr.predict(X), type='time_periods')}"
    )

    # Kendall correlation
    print(
        "\nKendall correlation over panel: "
        f"{correlation_coefficient(y, lr.predict(X), correlation_type='kendall', type='panel')}"
    )
    print(
        "Cross-sectional Kendall correlation: "
        f"{correlation_coefficient(y, lr.predict(X), correlation_type='kendall', type='cross_section')}"
    )
    print(
        "Periodic Kendall correlation: "
        f"{correlation_coefficient(y, lr.predict(X), correlation_type='kendall', type='time_periods')}"
    )
